from __future__ import print_function, division
from builtins import zip
from builtins import object
import numpy as np
from os import path
import logging
from itertools import groupby
import math

from virgotools.CfgParser import CfgParser
from virgotools.pvt_utils import add_property, check_frozen_setattr

__all__ = ['AclServer']


logger = logging.getLogger(__name__)

ACL_syntax = """
ACL_CONST_CH name unit ramptime:float fsample:float value:float
ACL_SUM_CH name unit ramptime:float *weights:float *inputs

ACL_RELAY_CH name unit fsample:float t01:float t10:float state:int input=None th01=0:float th10=0:float cond=>=

ACL_MATRIX_BEGIN name ramptime:float *inputs
ACL_MATRIX_CH name unit *weights:float
ACL_MATRIX_END name n_input=0:int

ACL_SINEWAVE_CH name unit ramptime:float fsample:float amplitude:float freq:float phi:float bias:float
ACL_SINELINE_CH name unit fsample:float freq:float phi=0:float

ACL_ADC_CH name unit ramptime:float fsample:float input bias:float slope:float filter

ACL_FILTER_CH name unit ramptime:float fsample:float input slope:float filter
"""

ACL_ignored_keys = """
CFG_PWD
CFG_PRIO
ACL_MAIN_LX
ACL_TIMING
ACL_INFO_OPTION
ACL_SMS
ACL_FILTER_SET
ACL_FILTER_POLES
ACL_FILTER_ZEROS
ACL_FILTER_BUTTERWORTH
ACL_OP_CH
ACL_DECL_CH
ACL_SWITCH_CH
ACL_CLIP_CH
ACL_FILTER_CH_RESET_CND
ACL_DEMOD
ACL_CHBUF_F
ACL_CHBUF_D
ACL_PERF
ACL_INFO
ACL_DAQ
ACL_DAQ_FILTER
ACL_NOISE_CH
ACL_SUM_CH_OPTION
ACL_FILTER_DEFAULT_MAX_ORDER
ACL_FILTER_CH_OPTION
TOLM_PROCESSOR_OUTPUT_PACKET_WITH_FREQUENCY
TOLM_PROCESSOR_PACKET_APPEND
"""
ACL_ignored_keys = set(ACL_ignored_keys.split())


notimplemented_keywords = set()

def acl_float(s):
    try:  # shortcut
        return float(s)
    except ValueError:
        pass
    if not s:
        raise ValueError('cannot convert empty string to float')
    ss = s.replace('PI', str(math.pi))
    if any(c not in ' 0123456789Ee+-/*.' for c in ss):
        raise ValueError('Illegal character in %r' % ss)
    try:
        r = eval('float(%s)' % ss)
        logger.debug('Floating point literal %r converted to %g', s, r)
        return r
    except BaseException:
        raise ValueError('Could not convert %r to float' % s)

acl_converters = {'float': acl_float, 'int': int}

AclParser = CfgParser(ACL_syntax, ACL_ignored_keys, converters=acl_converters)


# see http://stackoverflow.com/a/17576095
class SetOnly(object):
    def __init__(self, func):
        self.func = func

    def __get__(self, obj, objtype=None):
        raise NotImplementedError(
            'attribute can only be set, use AcParamGet to get value')

    def __set__(self, obj, value):
        return self.func(obj, value)


def make_cm(cm_type, params):
    """
    helper function to implement standard Cm methods with proper type conversion
    and useful docstring
    """
    arg_types = [param[1] for param in params]

    def fun(self, *args):  # captures cm_type, arg_types
        if len(args) > len(arg_types):
            raise ValueError('got too many parameters')

        # convert all args to required type. Missing arguments are skipped
        args = [arg_type(arg) for arg, arg_type in zip(args, arg_types)]

        # send cm_message to corresponding server and channel
        self._server.cm_send(cm_type, self._line.name, *args)

    # make helpful docstring
    fun.__doc__ = '{}({})'.format(
        cm_type, ', '.join('{}: {}'.format(param[0], param[1].__name__)
                           for param in params))
    return fun


class Base(object):
    _keep = True  # if block should be added for user

    def __init__(self, server, line):
        self._server = server
        self._line = line

    def AcParamGet(self):
        from virgotools.cm_lib import cm_send_receive  # only when needed
        reply = cm_send_receive('AcParam', 2, self._server._cm_name,
                                'AcParamGet', self._line.name)
        # just print for now, could potentially parse output
        print(reply[1])

    def __repr__(self):
        return repr(self._line)


class ACL_CONST_CH(Base):
    AcConstChSet = make_cm(
        'AcConstChSet', [('value', float), ('ramptime', float)])

    @SetOnly
    def value(self, value):
        self.AcConstChSet(value)

    def __set__(self, instance, value):
        self.AcConstChSet(value)


class SumInput(object):
    def __init__(self, input_):
        self._input = input_

    def __get__(self, obj, objtype=None):
        raise NotImplementedError(
            'attribute can only be set, use AcParamGet to see current value')

    def __set__(self, acl_sum, weight):
        acl_sum.AcSumChSet(acl_sum._line.ramptime, weight, self._input)


class ACL_SUM_CH(Base):
    def __init__(self, server, line):
        super(ACL_SUM_CH, self).__init__(server, line)

        for input_ in self._line.inputs:
            add_property(self, input_, SumInput(input_))
        self._isfrozen = True

    def __set__(self, instance, value):
        raise AttributeError("Cannot set ACL_SUM_CH, use input names as attributes")

    __setattr__ = check_frozen_setattr("ACL_SUM_CH has no input {}")

    def AcSumChSet(self, ramptime, *args):
        """AcSumChSet(ramptime, weight1, input1[, weight2, input2, ...]) """

        if len(args) % 2:
            raise ValueError('Need pairs of weight and input')

        weights_inputs = []
        for weight, input_ in zip(args[::2], args[1::2]):
            weights_inputs.append(float(weight))

            if input_ not in self._line.inputs:
                raise ValueError('Unknown input %s' % input_)
            weights_inputs.append(input_)

        self._server.cm_send('AcSumChSet', self._line.name, float(ramptime),
                             *weights_inputs)

    def reset(self, ramptime=None):
        """set weights of all inputs to zero"""
        args = []
        for input_ in self._line.inputs:
            args.extend([0, input_])
        self.AcSumChSet(self._line.ramptime, *args)


class ACL_ADC_CH(Base):
    AcAdcChSet = make_cm('AcAdcChSet', [
        ('bias', float),
        ('slope', float),
        ('ramptime', float),
        ('filter', str),
        ('filter_ramptime', float)])

    AcAdcChBiasSet = make_cm('AcAdcChBiasSet', [
        ('bias', float),
        ('ramptime', float)])

    AcAdcChSlopeSet = make_cm('AcAdcChSlopeSet', [
        ('slope', float),
        ('ramptime', float)])


class ACL_SINEWAVE_CH(Base):
    AcSinChSet = make_cm(
        'AcSinChSet',
        [('freq', float), ('amp', float), ('bias', float), ('phi', float),
         ('ramptime', float)])

    AcSinChFreqSet = make_cm(
        'AcSinChFreqSet',
        [('freq', float)])

    @SetOnly
    def freq(self, value):
        self.AcSinChFreqSet(value)

    AcSinChAmplSet = make_cm(
        'AcSinChAmplSet',
        [('amp', float), ('ramptime', float)])

    @SetOnly
    def amp(self, value):
        self.AcSinChAmplSet(value)


class ACL_SINELINE_CH(Base):
    AcLineChFreqSet= make_cm(
        'AcLineChFreqSet',
        [('freq', float)])

    @SetOnly
    def freq(self, value):
        self.AcLineChFreqSet(value)


class ACL_MATRIX_BEGIN(Base):
    def __init__(self, server, line):
        super(ACL_MATRIX_BEGIN, self).__init__(server, line)
        self.outputs = []
        self.values = []  # the matrix with weights
        assert server._current_matrix is None
        server._current_matrix = self

    def AcMatrixSet(self, *args):
        # too complex to enforce argument types?
        self._server.cm_send('AcMatrixSet', self._line.name, *args)

    def __getitem__(self, idx):
        # only gets values from original config file for now
        return self.values[idx]

    def __setitem__(self, idx, value):
        irows, icols = np.indices(self.values.shape)
        b = sorted(np.broadcast(irows[idx], icols[idx], value))
        b.sort()  # items with same row should be adjacent

        args = [self._line.ramptime]
        did_line = False
        for irow, group in groupby(b, lambda t: t[0]):  # group by irow
            if did_line:
                args.append('next_line')
            args.append(self.outputs[irow])
            for _, icol, val in group:
                args.append(float(val))
                args.append(self._line.inputs[icol])
            did_line = True

        self.AcMatrixSet(*args)

    def reset(self):
        """set weights of all inputs to zero"""
        self[:] = 0


class MatrixInput(object):
    def __init__(self, input_):
        self._input = input_

    def __get__(self, obj, objtype=None):
        raise NotImplementedError(
            'attribute can only be set, use AcParamGet to get value')

    def __set__(self, matrix_chan, weight):
        matrix_chan.AcMatrixChSet(matrix_chan._matrix._line.ramptime,
                                  weight, self._input)


class ACL_MATRIX_CH(Base):
    def __init__(self, server, line):
        super(ACL_MATRIX_CH, self).__init__(server, line)
        matrix = server._current_matrix
        self._matrix = matrix
        matrix.outputs.append(line.name)
        matrix.values.append(line.weights)

        for input_ in matrix._line.inputs:
            add_property(self, input_, MatrixInput(input_))

        self._isfrozen = True

    def AcMatrixChSet(self, ramptime, *args):
        """AcMatrixChSet(ramptime, weight1, input1[, weight2, input2, ...]) """
        if len(args) % 2:
            raise ValueError('Need pairs of weight and input')

        weights_inputs = []
        for weight, input_ in zip(args[::2], args[1::2]):
            weights_inputs.append(float(weight))

            if input_ not in self._matrix._line.inputs:
                raise ValueError('Unknown input %s' % input_)
            weights_inputs.append(input_)

        self._server.cm_send('AcMatrixChSet', self._line.name, ramptime,
                             *weights_inputs)


class ACL_MATRIX_END(Base):
    _keep = False

    def __init__(self, server, line):
        assert server._current_matrix._line.name == line.name
        server._current_matrix.values = np.array(server._current_matrix.values)
        server._current_matrix = None


class ACL_RELAY_CH(Base):

    AcRelayChSet = make_cm(
        'AcRelayChSet',
        [('time01', float), ('time10', float), ('start_value', int),
         ('thr01', float), ('thr10', float)])

    AcRelayChTranSet = make_cm('AcRelayChTranSet', [('value', int)])

    def __set__(self, instance, value):  # TODO: change to block.value?
        self.AcRelayChTranSet(value)


class ACL_FILTER_CH(Base):
    AcFltChFilterSet = make_cm(
        'AcFltChFilterSet', [('filter', str), ('ramptime', float)])

    @SetOnly
    def filter(self, filter):
        self.AcFltChFilterSet(filter)


class AclServer(object):
    """
    Pythonic interface to set parameters of an Acl server via Cm commands. Note
    that this is only useful for setting parameters. The current value of a
    parameter can be obtained by making it available in the frames (either
    directly as a fast channel, or as a slow channel using ACL_SMS).

    Attributes
    ----------
    acl_file : str
        name of config file of an Acl server. If no file extension is given, it is
        assumed to be the cm-name of a process, and the configuration file is
        assumed to be located in /virgoData/VirgoOnline

    Example
    -------
    >>> LSC = AclServer('LSC_Acl', ramptime=5)
    >>> LSC.LOCK_ENBL = 1
    cm_send('LSC_Acl', 'AcRelayChTranSet', 'LOCK_ENBL', 1, 5.0)

    """
    def __init__(self, acl_file, ramptime=1.0):
        if not acl_file.endswith('.cfg'):
            acl_file = '/virgoData/VirgoOnline/%s.cfg' % acl_file
        self._cm_name = path.splitext(path.basename(acl_file))[0]
        self._current_matrix = None  # for keeping track of begin_ .. end_matrix

        for line in AclParser.parse_file(acl_file):

            # TODO: implement repeated sums or fix config files
            if hasattr(self, line.name) and line.key not in ('ACL_MATRIX_END',):
                logger.warn('%s %s does already exist!',
                            line.key, line.name)

            try:
                cls = globals()[line.key]
            except KeyError:
                if line.key not in notimplemented_keywords:
                    logger.debug('Keyword %s not yet implemented', line.key)
                    notimplemented_keywords.add(line.key)
                continue

            obj = cls(self, line)

            if obj._keep:
                add_property(self, line.name, obj)
                #setattr(self, obj.name, obj)
        self._load = False
        self._isfrozen = True

    __setattr__ = check_frozen_setattr(
        "cannot assign to non-existing channel {}")

    def __repr__(self):
        return 'object for interfacing with Acl process %s' % self._cm_name

    def cm_send(self,  cm_type, *args):
        """same as cm_send, but targetted at this Acl server"""
        from virgotools.cm_lib import cm_send  # only when needed
        if self._load:
            args += ('load',)
        cm_send(self._cm_name, cm_type, *args)

    # context manager for sending commands at the same time
    def __enter__(self):
        logger.debug('Start loading commands')
        self._load = True
        return self  # not needed?

    def __exit__(self, exc_type, exc_val, exc_tb):
        logger.debug('Applying all loaded commands')
        self._load = False
        self.cm_send('AcParamSet')