"""
-------------
interface_tango
-------------
Pythonic interface for getting/setting parameters of various tango servers
"""
# B. Swinkels, 2016

from __future__ import division
from builtins import object
import os
import logging
from virgotools.pvt_utils import add_property, check_frozen_setattr


__all__ = ['TangoDevice']

logger = logging.getLogger(__name__)


# FIXME: allow giving shortcut name instead of full address

alias = {
    'Chiller': ('cmplvtf.virgo.infn.it:10000', 'INJ/LaserChiller/EER'),
    'DDS1': ('cmplvtf.virgo.infn.it:10000', 'INJ/sdt2011/DDS1'),
    'DDS2': ('cmplvtf.virgo.infn.it:10000', 'INJ/sdt2011/DDS2'),
    'SLXX': ('cmplvtf.virgo.infn.it:10000', 'INJ/SL_TempController/EER'),
    'LNFS': ('cmplvtf.virgo.infn.it:10000', 'INJ/Lnfs100/LNFS'),
}


class TangoAttribute(object):
    def __init__(self, attribute):
        self._attribute = attribute

    def __get__(self, instance, objtype=None):
        return instance._dev.read_attribute(self._attribute).value

    def __set__(self, instance, value):
        logger.info('Setting attribute %s of %s to %g',
                    self._attribute, instance._dev.name(), value)
        instance._dev.write_attribute(self._attribute, value)


class TangoDevice(object):
    """
    Class for interfacing with Tango devices, e.g. the ones for the LNFS.
    Interface is similar to the ones for the DSPs, but without the ramping.
    
    Parameters
    ----------
    tango_host : str
        address and port of the Tango host to which the server is connected
    device_name : str 
        name of the device
    
    Example
    -------
    >>> LNFS = TangoDevice('cmplvtf.virgo.infn.it:10000', 'INJ/Lnfs100/LNFS')
    >>> LNFS.FREQ1
    6270777.0
    """

    def __init__(self, tango_host, device_name):
        import PyTango  # only when needed

        os.environ['TANGO_HOST'] = tango_host  # fixme what happens when mixing tango servers on differnt hosts?
        self._dev = PyTango.DeviceProxy(device_name)

        for attr_name in self._dev.get_attribute_list():
            add_property(self, attr_name, TangoAttribute(attr_name))
        self._isfrozen = True  # protect against typos of attribute name

    __setattr__ = check_frozen_setattr()
