"""
-------------
virgo_logging
-------------

This module contains logging handlers for logging to console, directly to file
or via the Cfg library. These handlers are based on the
:py:mod:`logging` module from Python's standard library. A logger object called
*log* is created that can be used to for writing the actual logging messages.

Note: the logging handlers must typically installed at the top of your script
since Python will install some default handlers in case you start logging before
configuring anything.

Examples
--------
At the top of your script, install a logging handler in one of the following 3
ways:

>>> log_to_console(level='DEBUG')  # when debugging things on the command line
>>> log_to_file('/virgoLog/MyProcess/MyProcess')
# logs to file /virgoLog/MyProcess/MyProcess_2016-07-02-12h04m04-UTC.log
>>> log_to_cfg()  # for processes using FdIO

Logging messages can now be written using the *log* object

>>> logger.debug('Just testing ...')
>>> logger.info('This is interesting')
>>> logger.warn('Are you sure?')
>>> logger.error('Something is wrong')
"""

# B. Swinkels, 2014

from __future__ import division
from __future__ import print_function
from builtins import object
import logging
import time
import sys
import datetime as dt
from os import path
import subprocess as sp
from virgotools.pvt_utils import ascii


__all__ = ['log_to_console', 'log_to_file', 'log_to_cfg', 'log_command_line',
           'sec2str', 'wait', 'LoopTimer', 'fatal_exit', 'svn_version',
           'log_virgotools_version',]


logger = logging.getLogger(__name__)
datefmt = '%Y-%m-%d-%Hh%Mm%S-UTC' # see VIR-0256A-14
root = logging.getLogger()
root.setLevel(logging.DEBUG)  # overall level, set to DEBUG in case of trouble


def log_to_console(level='INFO'):
    """
    log messages to console

    Parameters
    ----------
    level : str or int
        logging level defined in the :py:mod:`logging` module, such as 'DEBUG',
        'INFO', 'WARN', 'ERROR'

    Example
    -------
    >>> log_to_console()
    >>> logger.debug('Some debug info')  # not displayed, level below threshold
    >>> logger.info('Some information')
    Some information
    """

    # prevent installing more than one handler when repeatedly running script
    # in ipython. Note that FileHandler is a subclass of StreamHandler, so
    # cannot test with isinstance.
    if not any(type(h) is logging.StreamHandler for h in root.handlers):
        handler = logging.StreamHandler()
        handler.setLevel(level)
        formatter = logging.Formatter('%(message)s')
        handler.setFormatter(formatter)
        root.addHandler(handler)


def log_to_file(log_file=None, level=logging.INFO):
    """
    log messages to file using Virgo format

    Parameters
    ----------
    log_file : str
        path to the log-file including directory and file-name prefix. A time
        stamp is added to the file name. The directory must already exist.
    level : str or int
        logging level defined in the :py:mod:`logging` module, such as 'DEBUG',
        'INFO', 'WARN', 'ERROR'

    Example
    -------
    >>> log_to_file('/virgoLog/MyTest')
    Logging to file /virgoLog/MyTest_2016-07-03-13h41m23-UTC.log
    >>> logger.debug('Some debug info')  # not written, level below threshold
    >>> logger.info('Some information')
    >>> logger.warn('Watch out!')
    >>> logger.error('This is bad!')

    Which is written to file as
    ::

        2016-07-03-13h41m43-UTC INFO  Some information
        2016-07-03-13h42m01-UTC WARNING Watch out!
        2016-07-03-13h42m08-UTC ERROR This is bad!
    """

    # default to just the name of the script
    if log_file is None:
        log_file = path.basename(path.splitext(sys.argv[0])[0])

    # FIXME: error when directory does not exist, create?
    log_file = '%s_%s.log' % (
        log_file, dt.datetime.utcnow().strftime(datefmt))
    logger.info('Logging to file %s', log_file)
    handler = logging.FileHandler(log_file, mode='w')
    handler.setLevel(level)
    formatter = logging.Formatter(
        fmt='%(asctime)s %(levelname)-5s %(message)s',
        datefmt=datefmt)
    formatter.converter = time.gmtime  # UTC
    handler.setFormatter(formatter)
    root.addHandler(handler)


'''
mapping from Python logging levels to Cfg logging levels:
    debug -> info, debug does not exist
    info -> info
    warn -> warning
    error -> error
    fatal -> error, want to do shut-down in python code
'''


class CfgHandler(logging.Handler):
    """logging handler that forwards messages to Virgo logging system"""
    def emit(self, record):
        import PyFd as fd  # only import when needed

        msg = self.format(record)
        msg = ascii(msg)
        if record.levelno <= logging.INFO:
            error_level = fd.CFG_INFO
        elif record.levelno <= logging.WARNING:
            error_level = fd.CFG_WARNING
        else:
            error_level = fd.CFG_ERROR
        fd.CfgMsgAddFmt(error_level, 0, msg)


def log_to_cfg(level=logging.INFO):
    """logs all messages to Virgo Cfg logging system. Note that any messages
    that are logged before FdIO is configured will be logged to the command line
    and not to file."""
    handler = CfgHandler()
    handler.setLevel(level)
    formatter = logging.Formatter(fmt='%(message)s')
    handler.setFormatter(formatter)
    root.addHandler(handler)


def log_command_line():
    """
    logs the command line of a script (as given by sys.argv). Useful to call as the very
    first thing after setting up loggers, to record exactly how a script was called.
    """
    import subprocess
    cmdline = subprocess.list2cmdline(sys.argv)
    logger.info('Started with command line:\n%s', cmdline)
    return cmdline


def svn_version(dirname):
    version = sp.check_output(['svnversion', dirname])
    logger.info('svnversion %s: %s', dirname, version)
    return version


def log_virgotools_version():
    pvt_dir = path.dirname(path.abspath(__file__))
    logger.info('Running PythonVirgoTools from directory %s', pvt_dir)
    svn_version(pvt_dir)


def parse_log_line(line):  #FIXME: test
    """parse time from a line in a standard Virgo logfile, and return a
    datetime object. Returns None if no valid time is detected"""
    try:
        t = dt.datetime.strptime(line[:23], datefmt)
    except ValueError:
        t = None
    return t


def parse_filename(filename):  #FIXME: test
    """splits filename of standard Virgo logfile in cm_name and datetime"""
    filename = path.basename(filename)
    filename, ext = path.splitext(filename)
    assert ext == '.log'
    cm_name, t = filename.rsplit('_', 1)
    t = dt.datetime.strptime(t, datefmt)
    return cm_name, t


def sec2str(s):
    """converts time in seconds to string"""
    m, s = divmod(s, 60)
    h, m = divmod(m, 60)
    if h:
        return '%i:%02i:%02i hour' % (h, m, s)
    if m:
        return '%i:%02i min' % (m, s)
    if s >= 10:
        return '%i second' % s
    if s >= 1:
        return '%.1f second' % s
    return '%.2f second' % s


class LoopTimer(object):
    """logs loop statistics

    Example
    -------
    >>> import time
    >>> log_to_console()
    >>> n = 3
    >>> timer = LoopTimer(n, 'calculating')
    calculating: Starting 3 loops
    >>> for i in range(3):
    ...     time.sleep(10)
    ...     timer.end(i)
    ...
    calculating: Finished loop 1 of 3, 30 second left
    calculating: Finished loop 2 of 3, 12 second left
    calculating: Finished 3 loops in 35 second, 11 second per loop.
    """
    def __init__(self, nloop, prefix='', tmin=0):
        """to be called before the start of the loop"""
        self.nloop = nloop
        self.tmin = tmin
        self.prefix = prefix

        self.t0 = time.time()
        self.tnext = self.t0 + self.tmin
        if self.prefix:
            self.prefix += ': '
        logger.info('%sStarting %i loops', self.prefix, nloop)

    def end(self, iloop):
        """to be called at end of loop"""
        now = time.time()
        dt = now - self.t0
        if iloop + 1 >= self.nloop:  # always print when finished
            logger.info('%sFinished %i loops in %s, %s per loop.',
                        self.prefix, iloop + 1, sec2str(dt), sec2str(dt / (iloop + 1)))
        elif now >= self.tnext:  # only print when not done too recent
            self.tnext = now + self.tmin
            eta = (self.nloop - (iloop+1)) * dt / (iloop+1)
            logger.info('%sFinished loop %i of %i, %s left',
                        self.prefix, iloop + 1, self.nloop, sec2str(eta))


def fatal_exit(*args, **kwargs):
    if args or kwargs:
        logger.critical(*args, **kwargs)
    logger.critical('Fatal error, shutting down')
    sys.exit(1)


def wait(dur, msg=None, step=60):
    """Waits given period in seconds. Logs optional message"""
    if msg:
        logger.info("Waiting %s for %s", sec2str(dur), msg)
    else:
        logger.info("Waiting %s", sec2str(dur))

    while dur > step:
        time.sleep(step)
        dur -= step
        print('still waiting %s ...' % sec2str(dur))
        sys.stdout.flush()

    time.sleep(dur)
