from __future__ import print_function
import os
import pytest
import mock

from virgotools import cfg_lib
from virgotools.cfg_lib import strip_multiline_comments, parse_config_iter, parse_key



# mock file that allow iteration, see last example:
# https://stackoverflow.com/a/24779923



# hack to make it work on both Py2 and Py3:
try:
    import __builtin__
    OPEN_CMD = '__builtin__.open'
    print('Python2')
except:
    OPEN_CMD = 'builtins.open'
    print('Python3')



def mock_open_fs(filesystem):
    """mock open that allows to iterate over text files from fake file system given as
    a dict fname -> txt"""

    def my_open(fname):
        try:
            txt = filesystem[fname]
        except KeyError:
            raise KeyError('got asked %s, only got %s' % (fname, filesystem.keys()))
        # mock file that allows iteration, see last example:
        # https://stackoverflow.com/a/24779923
        m = mock.mock_open(read_data=txt)
        m.return_value.__iter__ = lambda self: iter(self.readline, '')
        return m()
    return my_open


multiline = """\
line/* inline comment*/ one
line two
/*
multiple
lines
*/
line seven
before/* remaining stuff inside 
multiline comment
firts part inside */ after
last line
""".splitlines(True)


def test_multiline():
    actual = list(strip_multiline_comments(multiline))
    expected = [(1, 'line one\n'), (2, 'line two\n'), (3, '\n'),
                (7, 'line seven\n'), (8, 'before after\n'), (11, 'last line\n')]
    assert actual == expected


not_closed = """\
line one
line two/* this is a
multiline comment
that is never closed
""".splitlines(True)


def test_comment_not_closed():
    # could test that message is logged, but don't bother
    assert list(strip_multiline_comments(not_closed)) == [
        (1, 'line one\n'), (2, 'line two')]


fs = {'/cwd/a.cfg': '''\
key1 5
'''}

CWD = '/cwd'

def test_yy(monkeypatch):
    monkeypatch.setattr(OPEN_CMD, mock_open_fs(fs))
    monkeypatch.setattr(os, 'getcwd', mock.Mock(return_value=CWD))

    actual = list(parse_config_iter('a.cfg'))
    expected = [('/cwd/a.cfg', 1, ['key1', '5'])]
    assert actual == expected


fs_include = {
    '/cwd/main.cfg': '''\
KEY1 before include
include /absolute/path/sub1.cfg
include relative/path/sub2.cfg
KEY4 after include
''',
    '/absolute/path/sub1.cfg': '''\
KEY2 a b c    
             ''',
    '/cwd/relative/path/sub2.cfg':'''\
KEY3 1 2 3    
             ''',}

def test_include(monkeypatch):
    monkeypatch.setattr(OPEN_CMD, mock_open_fs(fs_include))
    monkeypatch.setattr(os, 'getcwd', mock.Mock(return_value=CWD))

    actual = list(parse_config_iter('main.cfg'))
    expected = [
        ('/cwd/main.cfg', 1, ['KEY1', 'before', 'include']),
        ('/absolute/path/sub1.cfg', 1, ['KEY2', 'a', 'b', 'c']),
      ('/cwd/relative/path/sub2.cfg', 1, ['KEY3', '1', '2', '3']),
    ('/cwd/main.cfg', 4, ['KEY4', 'after', 'include'])]
    assert actual == expected


fs_define = {
    '/cwd/main.cfg': '''\
define alias 55
KEY1 alias

define chain1 abc
define chain2 chain1
KEY2 chain2
'''}


def test_define(monkeypatch):
    monkeypatch.setattr(OPEN_CMD, mock_open_fs(fs_define))
    monkeypatch.setattr(os, 'getcwd', mock.Mock(return_value=CWD))

    actual = list(parse_config_iter('main.cfg'))
    expected = [
        ('/cwd/main.cfg', 2, ['KEY1', '55']),
        ('/cwd/main.cfg', 6, ['KEY2', 'abc']),
    ]
    assert actual == expected


fs_continuation = {
    '/cwd/main.cfg':
r'''KEY1 a b c
KEY2 d e \
f g
KEY3 "continuation within \
a string"
KEY4 h i j 
'''}


def test_continuation(monkeypatch):
    monkeypatch.setattr(OPEN_CMD, mock_open_fs(fs_continuation))
    monkeypatch.setattr(os, 'getcwd', mock.Mock(return_value=CWD))

    actual = list(parse_config_iter('main.cfg'))
    expected = [
        ('/cwd/main.cfg', 1, ['KEY1', 'a', 'b', 'c']),
        ('/cwd/main.cfg', 2, ['KEY2', 'd', 'e', 'f', 'g']),
        ('/cwd/main.cfg', 4, ['KEY3', 'continuation within a string']),
        ('/cwd/main.cfg', 6, ['KEY4', 'h', 'i', 'j']),
    ]
    assert actual == expected


fs3 = {'/cwd/parse_key.cfg': """\
KEY
OTHER_KEY
KEY a
KEY b c
# KEY commented
KEY "with spaces"
"""}


def test_parse_key(monkeypatch):
    monkeypatch.setattr(OPEN_CMD, mock_open_fs(fs3))
    monkeypatch.setattr(os, 'getcwd', mock.Mock(return_value='/cwd/'))

    assert parse_key('parse_key.cfg', 'KEY') == [[], ['a'], ['b', 'c'], ['with spaces']]


def test_comments():
    pass